import 'dart:async';

import 'package:rxdart/rxdart.dart';
import 'package:bloc/bloc.dart' as bloc;

import '../state_graph_bloc.dart';
import 'graph.dart';

/// Base class for Blocs using [StateGraph], based on bloc package.
abstract class StateGraphBloc<Event, State> extends bloc.Bloc<Event, State> {
  final CompositeSubscription _compositeSubscription = CompositeSubscription();

  StateGraph<Event, State> _graph;

  /// Method for building [StateGraph] instance.
  StateGraph<Event, State> buildGraph();

  @override
  Stream<State> mapEventToState(Event event) {
    final graph = _getOrCreateGraph();
    final transitionEntry = graph.findTransitionEntry(state, event);
    if (transitionEntry == null) return Stream.value(state);

    final oldState = state;
    final newState = transitionEntry.transition?.call(state, event) ?? state;
    transitionEntry?.sideEffect?.call(oldState, event);
    return Stream.value(newState);
  }

  /// Maps current state stream to the new one using [convert] method.
  /// Uses [distinct] operator to skip recurring values.
  Stream<N> bindState<N>(N convert(State currentState)) =>
      map(convert).distinct();

  StateGraph<Event, State> _getOrCreateGraph() {
    if (_graph == null) {
      _graph = buildGraph();
      assert(_graph != null);
    }
    return _graph;
  }

  /// Launch [action] subscription that will be closed on Bloc's [close] method call.
  /// Converts [action] to [Stream] and returns [Future] which is exactly [launchStream] subscription.
  StreamSubscription<T> launchFuture<T>(
    Future<T> action, {
    void onData(T event),
    Function onError,
    void onDone(),
    bool cancelOnError,
  }) {
    final stream = action.asStream();
    return launchStream(
      stream,
      onData: onData,
      onError: onError,
      onDone: onDone,
      cancelOnError: cancelOnError,
    );
  }

  /// Launch [action] subscription that will be cancelled on Bloc's [close] method call.
  /// Returns immediately cancelled [StreamSubscription] if bloc was already disposed.
  StreamSubscription<T> launchStream<T>(
    Stream<T> action, {
    void onData(T event),
    Function onError,
    void onDone(),
    bool cancelOnError,
  }) {
    if (_compositeSubscription.isDisposed) {
      return _EmptyStreamSubscription<T>();
    }

    final subscription = action.listen(
      onData ?? (_) {},
      onError: onError,
      onDone: onDone,
      cancelOnError: cancelOnError,
    );

    try {
      return _compositeSubscription.add(subscription);
    } on String {
      // Composite subscription was disposed in the meantime.
      subscription.cancel();
      return _EmptyStreamSubscription<T>();
    }
  }

  /// Builds subject for [SingleLiveEvent]s.
  /// Is auto-closed on [close] method.
  SingleLiveEventSubject<T> singleLiveEventSubject<T>() {
    final result = _AutoCloseSingleLiveEventSubject<T>();
    _compositeSubscription.add(result);
    return result;
  }

  @override
  Future<void> close() async {
    _compositeSubscription.dispose();
    await super.close();
  }
}

class _AutoCloseSingleLiveEventSubject<T> extends StreamSubscription<T>
    with _EmptyStreamSubscription<T>, SingleLiveEventSubject<T> {
  final _subject = BehaviorSubject<SingleLiveEvent<T>>();

  @override
  Stream<SingleLiveEvent<T>> get stream => _subject;

  @override
  void add(T data) {
    _subject.add(SingleLiveEvent(data));
  }

  @override
  void close() {
    _subject.close();
  }

  @override
  Future cancel() {
    super.cancel();
    close();
    return Future.value(null);
  }
}

class _EmptyStreamSubscription<T> implements StreamSubscription<T> {
  @override
  Future<E> asFuture<E>([E futureValue]) {
    return Future.value(futureValue);
  }

  @override
  Future cancel() {
    return Future.value(null);
  }

  @override
  bool get isPaused => false;

  @override
  void onData(void Function(T data) handleData) {}

  @override
  void onDone(void Function() handleDone) {}

  @override
  void onError(Function handleError) {}

  @override
  void pause([Future resumeSignal]) {}

  @override
  void resume() {}
}
